
open Wdigraph

let astar (g: wdigraph) (src: int) (dst: int) (h: int -> float) : float =
  let n = size g in
  let dist = Array.make n infinity in
  let pqueue = Pqueue.create () in
  let add v d =
    dist.(v) <- d; Pqueue.insert pqueue (d +. h v, v) in
  add src 0.;
  let relax v (w, dvw) =
    let d = dist.(v) +. dvw in
    if d < dist.(w) then add w d in
  let rec loop () =
    if Pqueue.is_empty pqueue then raise Not_found;
    let _, v = Pqueue.extract_min pqueue in
    if v = dst then
      dist.(dst)
    else (
      List.iter (relax v) (succ g v);
      loop ()
    ) in
  loop ()

(* reconstruit le chemin de src à v, avec path donnant le prédécesseur *)
let rec build_path path src acc v =
  if v = src then src :: acc else build_path path src (v :: acc) path.(v)

let astar_path g src dst h : float * int list =
  let n = size g in
  let dist = Array.make n infinity in
  let path = Array.make n (-1) in
  let pqueue = Pqueue.create () in
  let add o v d =
    dist.(v) <- d; path.(v) <- o; Pqueue.insert pqueue (d +. h v, v) in
  add src src 0.;
  let relax v (w, dvw) =
    let d = dist.(v) +. dvw in
    if d < dist.(w) then add v w d in
  let rec loop () =
    if Pqueue.is_empty pqueue then raise Not_found;
    let _, v = Pqueue.extract_min pqueue in
    if v = dst then
      dist.(dst), build_path path src [] dst
    else (
      List.iter (relax v) (succ g v);
      loop ()
    ) in
  loop ()

let astar_count g src dst h : int * float =
  let n = size g in
  let dist = Array.make n infinity in
  let pqueue = Pqueue.create () in
  let count = ref 0 in
  let add v d =
    incr count;
    dist.(v) <- d; Pqueue.insert pqueue (d +. h v, v) in
  add src 0.;
  let relax v (w, dvw) =
    let d = dist.(v) +. dvw in
    if d < dist.(w) then add w d in
  let rec loop () =
    if Pqueue.is_empty pqueue then raise Not_found;
    let _, v = Pqueue.extract_min pqueue in
    if v = dst then
      !count, dist.(dst)
    else (
      List.iter (relax v) (succ g v);
      loop ()
    ) in
  loop ()
