/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Tableau de bits
// (8 fois plus compact qu'un tableau de type bool[])

#include "prelude.h"
#include "bitarray.h"

struct Bitarray {
  int size;
  uint32_t *bits;
};

// W = nombre de bits par élément du tableau
#define W (8 * sizeof(uint32_t))

bitarray *bitarray_create(int size, bool b) {
  assert(size >= 0);
  bitarray *a = malloc(sizeof(struct Bitarray));
  a->size = size;
  int n = size / W;
  if (n % W > 0) n++;
  a->bits = calloc(n, sizeof(uint32_t));
  if (b)
    for (int i = 0; i < n; i++)
      a->bits[i] = ~0;
  return a;
}

int bitarray_size(bitarray *a) {
  return a->size;
}

void bitarray_set(bitarray *a, int i, bool b) {
  assert(0 <= i && i < a->size);
  int j = i / W, k = i % W;
  if (b)
    a->bits[j] |= 1 << k;
  else
    a->bits[j] &= ~(1 << k);
}

bool bitarray_get(bitarray *a, int i) {
  assert(0 <= i && i < a->size);
  int j = i / W, k = i % W;
  return (a->bits[j] & (1 << k)) != 0;
}

void bitarray_delete(bitarray *a) {
  free(a->bits);
  free(a);
}
