
type in_file = {
  cin: in_channel;
  mutable byte: int;
  mutable ofs: int; (* prochain bit à lire dans `byte` *)
                    (* -1 signifie qu'il faut lire un nouvel octet *)
}

let open_in ~file =
  { cin = open_in file; byte = 0; ofs = -1 }

let read_bit inf =
  if inf.ofs = -1 then (
    inf.byte <- Char.code (input_char inf.cin);
    inf.ofs <- 0
  );
  let b = if inf.byte land (1 lsl inf.ofs) > 0 then 1 else 0 in
  inf.ofs <- 1 + inf.ofs;
  if inf.ofs = 8 then inf.ofs <- -1;
  b

let close_in inf =
  close_in inf.cin

type out_file = {
  cout: out_channel;
  mutable byte: int; (* prochain bit à écrire *)
  mutable ofs: int; (* 0..7 ; prochaine position où écrire dans `byte` *)
}

let open_out ~file =
  { cout = open_out file; byte = 0; ofs = 0 }

let emit_bit outf =
  output_char outf.cout (Char.chr outf.byte);
  outf.byte <- 0;
  outf.ofs <- 0

let write_bit outf b =
  if b = 1 then outf.byte <- outf.byte lor (1 lsl outf.ofs);
  if outf.ofs = 7 then emit_bit outf else outf.ofs <- 1 + outf.ofs

let close_out outf =
  if outf.ofs > 0 then emit_bit outf;
  flush outf.cout;
  close_out outf.cout
