/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Filtre de Bloom pour des chaînes (char*)

#include "prelude.h"
#include <limits.h>
#include "bloom.h"

// une famille de fonctions de hachage
int bloom_hash(int seed, char *s) {
  int h = 0;
  char c;
  while ((c = *s++) != 0) {
    h = seed * h + c;
  }
  return h & INT_MAX;
}

struct Bloom {
  int k;
  int *seeds; // tableau de taille k
  int m;
  bool *bits; // tableau de taille m
};

bloom *bloom_create(int k, int m) {
  bloom *b = malloc(sizeof(struct Bloom));
  b->k = k;
  b->seeds = calloc(k, sizeof(int));
  for (int i = 0; i < k; i++) b->seeds[i] = rand();
  b->m = m;
  b->bits = calloc(m, sizeof(bool));
  return b;
}

int bloom_bit(bloom *b, int i, char *s) {
  return bloom_hash(b->seeds[i], s) % b->m;
}

void bloom_add(bloom *b, char *s) {
  for (int i = 0; i < b->k; i++) {
    b->bits[bloom_bit(b, i, s)] = true;
  }
}

bool bloom_contains(bloom *b, char *s) {
  for (int i = 0; i < b->k; i++) {
    if (!b->bits[bloom_bit(b, i, s)])
      return false;
  }
  return true;
}

void bloom_delete(bloom *b) {
  free(b->bits);
  free(b);
}
