
(* Implémentation d'un tableau associatif par un arbre binaire de
   recherche.

   La comparaison est faite sur la clé, en utilisant la comparaison
   polymorphe d'OCaml. *)

type ('k, 'v) bst =
  | E
  | N of ('k, 'v) bst * 'k * 'v * ('k, 'v) bst

let empty  : ('k, 'v) bst =
  E

let is_empty (t: ('k, 'v) bst) : bool =
  t = E

let rec mem k t = match t with
  | E ->
      false
  | N (l, k', _, r) ->
      if k < k' then mem k l
      else if k > k' then mem k r
      else true

let rec find (k: 'k) (t: ('k, 'v) bst) : 'v =
  match t with
  | E ->
      raise Not_found
  | N (l, k', v', r) ->
      if k < k' then find k l
      else if k > k' then find k r
      else v'

let rec add (k: 'k) (v: 'v) (t: ('k, 'v) bst) : ('k, 'v) bst =
  match t with
  | E ->
      N (E, k, v, E)
  | N (l, k', v', r) ->
      if k < k' then N (add k v l, k', v', r)
      else if k > k' then N (l, k', v', add k v r)
      else N (l, k, v, r) (* on écrase la valeur précédente *)

let rec min_elt = function
  | E ->
      raise Not_found
  | N (E, k, v, _) ->
      k, v
  | N (l, _, _, _) ->
      min_elt l

(* suppose l'arbre non vide *)
let rec remove_min_elt = function
  | E ->
      assert false
  | N (E, _, _, r) ->
      r
  | N (l, k, v, r) ->
      N (remove_min_elt l, k, v, r)

let rec remove k = function
  | E ->
      E
  | N (l, k', v', r) ->
      if k < k' then N (remove k l, k', v', r)
      else if k > k' then N (l, k', v', remove k r)
      else if r = E then l
      else let km, vm = min_elt r in N (l, km, vm, remove_min_elt r)

let rotate_right = function
  | N (N (t1, ku, vu, t2), kv, vv, t3) ->
      N (t1, ku, vu, N (t2, kv, vv, t3))
  | t -> t

let rotate_left = function
  | N (t1, ku, vu, N (t2, kv, vv, t3)) ->
      N (N (t1, ku, vu, t2), kv, vv, t3)
  | t -> t
