
(* Deux plus proches points du plan (diviser pour régner) *)

type point = float * float

let dist ((x1,y1), (x2,y2)) =
  sqrt ((x1-.x2) *. (x1-.x2) +. (y1-.y2) *. (y1-.y2))

let compare_x (x1, _) (x2, _) =
  compare x1 x2

let compare_y (_, y1) (_, y2) =
  compare y1 y2

(* calcul direct pour 2 ou 3 points seulement *)
let closest_pair_small (a: point array) : point * point =
  match a with
  | [| p; q |] ->
      (p, q)
  | [| p; q; r |] ->
      let pq = dist (p, q) in
      let qr = dist (q, r) in
      let pr = dist (p, r) in
      if pq < qr
      then if pr < pq then (p, r) else (p, q)
      else if pr < qr then (p, r) else (q, r)
  | _ ->
      assert false

let closest_pair_in_strip (a: point array) : point * point =
  let n = Array.length a in
  let d = ref (dist (a.(0), a.(1))) in
  let p = ref 0 in
  let q = ref 1 in
  for i = 0 to n - 1 do
    for j = i + 1 to min (n - 1) (i + 7) do
      let e = dist (a.(i), a.(j)) in
      if e < !d then (d := e; p := i; q := j)
    done
  done;
  (a.(!p), a.(!q))

let select (a: point array) (xmin: float) (xmax: float) : point array =
  let lst = ref [] in
  for i = Array.length a - 1 downto 0 do
    let x, _ = a.(i) in
    if xmin <= x && x <= xmax then lst := a.(i) :: !lst
  done;
  Array.of_list !lst

let closest_pair (a: point array) : point * point =
  if Array.length a < 2 then invalid_arg "closest_pair";
  let ax = Array.copy a in Array.sort compare_x ax;
  let ay = Array.copy a in Array.sort compare_y ay;
  let rec closest_rec lo hi =
    if hi - lo <= 3 then
      closest_pair_small (Array.sub ax lo (hi - lo))
    else (
      let mid = lo + (hi - lo) / 2 in
      let pql = closest_rec lo mid in
      let pqr = closest_rec mid hi in
      let pq = if dist pql < dist pqr then pql else pqr in
      let d = dist pq in
      let xmed = fst ax.(mid - 1) in
      let strip =
        select ay (max (xmed -. d) (fst ax.(lo    )))
                  (min (xmed +. d) (fst ax.(hi - 1))) in
      if Array.length strip <= 1 then
        pq
      else (
        let pqs = closest_pair_in_strip strip in
        if dist pqs < d then pqs else pq
      )
    )
  in
  closest_rec 0 (Array.length a)

(* version naïve, quadratique *)
let closest_pair_brute_force (a: point array) : point * point =
  if Array.length a < 2 then invalid_arg "closest_pair_brute_force";
  let n = Array.length a in
  let p = ref a.(0) in
  let q = ref a.(1) in
  let d = ref (dist (!p, !q)) in
  for i = 0 to n - 2 do
    for j = i + 1 to n - 1 do
      let e = dist (a.(i), a.(j)) in
      if e < !d then (
        d := e;
        p := a.(i);
        q := a.(j)
      )
    done
  done;
  (!p, !q)
