(* Réduction de kCOLOR vers SAT *)

open Cnf
open Dpll
open Graph
open Format

let color2sat g k =
  let n = size g in
  let var s c = k * s + c + 1 in
  let clauses = ref [] in
  (* tous les sommest sont colorés *)
  for s = 0 to n - 1 do
    clauses := List.init k (fun c -> var s c) :: !clauses
  done;
  (* chaque sommet a une unique couleur *)
  for s = 0 to n - 1 do
    for c1 = 0 to k - 2 do
      for c2 = c1 + 1 to k - 1 do
        clauses := [-var s c1; -var s c2] :: !clauses
      done
    done
  done;
  (* deux sommets adjacents ont des couleurs différentes *)
  List.iter (fun (s1, s2) ->
    for c = 0 to k - 1 do
      clauses := [-var s1 c; -var s2 c] :: !clauses
    done) (edges g);
  { kind = CNF; nbvars = n * k; clauses = !clauses }

let color_using_sat g k =
  let nf = color2sat g k in
  let n = size g in
  let var s c = k * s + c + 1 in
  match dpll nf with
  | None   -> None
  | Some v ->
     (* décodage de la coloration *)
     let color = Array.make n 0 in
     for s = 0 to n - 1 do
       for c = 0 to k - 1 do
         if v.(var s c) then color.(s) <- c
       done
     done;
     Some color

