/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// graphe orienté

// implémentation par listes d'adjacence

#include "prelude.h"
#include "list.h"
#include "digraph.h"

struct Digraph {
  int nbv;
  list **adj;
};

digraph *digraph_create(int size) {
  assert(size >= 0);
  digraph *g = malloc(sizeof(struct Digraph));
  g->nbv = size;
  g->adj = calloc(size, sizeof(list));
  for (int i = 0; i < size; i++) {
    g->adj[i] = NULL;
  }
  return g;
}
int digraph_size(digraph *g) {
  return g->nbv;
}
bool digraph_is_vertex(digraph *g, int u) {
  return 0 <= u && u < g->nbv;
}
bool digraph_has_edge(digraph *g, int u, int v) {
  assert(digraph_is_vertex(g, u) && digraph_is_vertex(g, v));
  for (list *l = g->adj[u]; l != NULL; l = l->next) {
    if (l->value == v) {
      return true;
    }
  }
  return false;
}
void digraph_add_edge(digraph *g, int u, int v) {
  assert(digraph_is_vertex(g, u) && digraph_is_vertex(g, v));
  if (!digraph_has_edge(g, u, v))
    g->adj[u] = list_cons(v, g->adj[u]);
}
list *digraph_succ(digraph *g, int u) {
  assert(digraph_is_vertex(g, u));
  return g->adj[u];
}
int digraph_nbe(digraph *g) {
  int e = 0;
  for (int i = 0; i < g->nbv; i++)
    for (list *l = g->adj[i]; l != NULL; l = l->next)
      e++;
  return e;
}
void digraph_delete(digraph *g) {
  for (int i = 0; i < g->nbv; i++) {
    list_delete(g->adj[i]);
  }
  free(g);
}

digraph *digraph_load(char *file) {
  FILE *f = fopen(file, "r");
  int n;
  if (fscanf(f, "%d\n", &n) == EOF) { fprintf(stderr, "error\n"); exit(1); }
  digraph *g = digraph_create(n);
  int e;
  if (fscanf(f, "%d\n", &e) == EOF) { fprintf(stderr, "error\n"); exit(1); }
  while (e-- > 0) {
    int i, j;
    if (fscanf(f, "%d %d\n", &i, &j) == EOF) { fprintf(stderr, "error\n"); exit(1); }
    digraph_add_edge(g, i, j);
  }
  fclose(f);
  return g;
}

void digraph_print(digraph *g) {
  printf("%d\n", g->nbv);
  int e = digraph_nbe(g);
  printf("%d\n", e);
  for (int i = 0; i < g->nbv; i++)
    for (list *l = g->adj[i]; l != NULL; l = l->next)
      printf("%d %d\n", i, l->value);
}

digraph *digraph_mirror(digraph *g) {
  digraph *m = digraph_create(g->nbv);
  for (int i = 0; i < g->nbv; i++)
    for (list *l = g->adj[i]; l != NULL; l = l->next)
      digraph_add_edge(m, l->value, i);
  return m;
}

int dfs_post(digraph *g, int order[], int next, bool visited[], int v) {
  if (visited[v]) return next;
  visited[v] = true;
  for (list *l = g->adj[v]; l != NULL; l = l->next)
    next = dfs_post(g, order, next, visited, l->value);
  order[next] = v;
  return next-1;
}

// remplit le tableau order avec l'ordre DFS postfixe
// suppose que order a la taille g->nbv
void digraph_postorder(digraph *g, int order[]) {
  bool *visited = calloc(g->nbv, sizeof(bool));
  for (int i = 0; i < g->nbv; i++) visited[i] = false;
  int next = g->nbv-1;
  for (int i = 0; i < g->nbv; i++)
    next = dfs_post(g, order, next, visited, i);
  assert(next == -1);
}
