
(* Graphe orienté.

   Les sommets sont les entiers 0,1,...,size-1.
*)

type digraph

val create: int -> digraph

val size: digraph -> int

val has_edge: digraph -> int -> int -> bool

val add_edge: digraph -> int -> int -> unit

val succ: digraph -> int -> int list

type edge = int * int

val edges: digraph -> edge list

val mirror: digraph -> digraph

val dfs: digraph -> int -> bool array
  (* Parcours en profondeur à partir d'une source donnée.
     Renvoie un tableau indiquant les sommets visités. *)

val dfs_stack: digraph -> int -> bool array
  (* Même chose, sans récursivité. *)

val exists_path: digraph -> int -> int -> bool

val dfs_path: digraph -> int -> int array
  (* Renvoie un tableau indiquant, pour chaque sommet atteint, son
     prédécesseur dans le parcours ; permet donc de reconstuire le chemin.
     Le prédécesseur de la source est elle-même. Les sommets non atteints
     ont `-1` comme prédécesseur. *)

val path: digraph -> int -> int -> int list
  (* `path g u v` renvoie un chemin de `u` à `v`, s'il en existe un,
     et lève l'exception `Not_found` sinon *)

val has_cycle: digraph -> int -> bool
  (* `has_cycle g s` détermine s'il existe un cycle accessible à partir de s *)

val bfs: digraph -> int -> int array
  (* Parcours en largeur à partir d'une source donnée.
     Renvoie un tableau indiquant la distance à la source,
     avec la valeur `max_int` pour les sommets non atteints. *)

val post_order : digraph -> int list
  (* Les sommets de `g` selon un ordre DFS postfixe. *)

val matrix: digraph -> int array array
  (* La matrice d'adjacence, par exemple pour faire des calculs de
     nombre de chemins. *)

val load: string -> digraph

val print: Format.formatter -> digraph -> unit

(* Quelques graphes particuliers *)

val cycle: int -> digraph
  (* Le cycle 0->1->2->...->(n-1). *)

val full: ?loop:bool -> int -> digraph
  (* Graphe complet, avec ou sans boucles. *)

val grid: int -> int -> digraph
  (* grille n*m où le sommet (i,j) est l'entier i*m+j,
     avec des arcs vers la droite et vers le bas *)
