
open Wdigraph

let dijkstra (g: wdigraph) (source: int) : float array =
  let n = size g in
  let dist = Array.make n infinity in
  let pqueue = Pqueue.create () in
  let visited = Array.make n false in
  let add v d = dist.(v) <- d; Pqueue.insert pqueue (d, v) in
  add source 0.;
  while not (Pqueue.is_empty pqueue) do
    let dv, v = Pqueue.extract_min pqueue in
    if not visited.(v) then (
      visited.(v) <- true;
      (* on vient de déterminer la distance de v *)
      List.iter
        (fun (w, dvw) ->
          let d = dv +. dvw in
          if d < dist.(w) then add w d)
        (succ g v)
    )
  done;
  dist

let dijkstra_path (g: wdigraph) (source: int) : float array * int array =
  let n = size g in
  let dist = Array.make n infinity in
  let pqueue = Pqueue.create () in
  let visited = Array.make n false in
  let path = Array.make n (-1) in
  let add o v d = dist.(v) <- d; Pqueue.insert pqueue (d, v); path.(v) <- o in
  add source source 0.;
  while not (Pqueue.is_empty pqueue) do
    let dv, v = Pqueue.extract_min pqueue in
    if not visited.(v) then (
      visited.(v) <- true;
      (* on vient de déterminer la distance de v *)
      List.iter
        (fun (w, dvw) ->
          let d = dv +. dvw in
          if d < dist.(w) then add v w d)
        (succ g v)
    )
  done;
  dist, path

let rec build_path path src acc v =
  if v = src then src :: acc else build_path path src (v :: acc) path.(v)

let dijkstra_dst g source dst =
  let n = size g in
  let dist = Array.make n infinity in
  let pqueue = Pqueue.create () in
  let visited = Array.make n false in
  let path = Array.make n (-1) in
  let count = ref 0 in
  let add o v d =
    incr count; dist.(v) <- d; Pqueue.insert pqueue (d, v); path.(v) <- o in
  add source source 0.;
  let rec loop () =
    if Pqueue.is_empty pqueue then raise Not_found;
    let dv, v = Pqueue.extract_min pqueue in
    if v = dst then
      !count, dv, build_path path source [] dst
    else (
      if not visited.(v) then (
        visited.(v) <- true;
        List.iter
          (fun (w, dvw) -> let d = dv +. dvw in if d < dist.(w) then add v w d)
          (succ g v)
      );
      loop ()
    )
  in
  loop ()
