
(* Graphes orientés pondérés par la distance euclidienne,
   par listes d'adjacence *)

type eucldigraph = {
  nbv: int;
  pos: (float * float) array;
  adj: int list array;
}

let create n f =
  { nbv = n; pos = Array.init n f; adj = Array.make n []; }

let size g =
  g.nbv

let pos g v =
  g.pos.(v)

let has_edge g u v =
  List.mem v g.adj.(u)

let add_edge g u v =
  assert (not (has_edge g u v));
  g.adj.(u) <- v :: g.adj.(u)

let distance (xu, yu) (xv, yv) =
  let dx = xu -. xv and dy = yu -. yv in
  sqrt (dx *. dx +. dy *. dy)

let weight g u v =
  assert (has_edge g u v);
  distance g.pos.(u) g.pos.(v)

let succ g u =
  g.adj.(u)

let wdigraph g =
  let wg = Wdigraph.create g.nbv in
  let add u v = Wdigraph.add_edge wg (weight g u v, u, v) in
  let add u vl = List.iter (add u) vl in
  Array.iteri add g.adj;
  wg

open Scanf

let load file =
  let c = Scanning.open_in file in
  let nbv = bscanf c "%d\n" (fun n -> n) in
  let pos = Array.make nbv (0., 0.) in
  for _ = 1 to nbv do
    bscanf c "%d %f %f\n" (fun u x y -> pos.(u) <- x, y)
  done;
  let g = { nbv; pos; adj = Array.make nbv [] } in
  let e = bscanf c "%d\n" (fun e -> e) in
  for _ = 1 to e do
    bscanf c "%d %d\n" (add_edge g)
  done;
  Scanning.close_in c;
  g

open Format

let print fmt g =
  fprintf fmt "%d@\n" g.nbv;
  for i = 0 to g.nbv - 1 do
    let x, y = g.pos.(i) in fprintf fmt "%d %f %f@\n" i x y
  done;
  let e = Array.fold_left (List.fold_left (fun e _ -> e+1)) 0 g.adj in
  fprintf fmt "%d@\n" e;
  for i = 0 to g.nbv - 1 do
    List.iter (fun j -> fprintf fmt "%d %d@\n" i j) g.adj.(i)
  done

