
(* Graphe orienté pondéré par la distance euclidienne

  Les sommets sont les entiers 0,1,...,size-1.
  La position de chaque sommet est spécifiée à la création.
*)

type eucldigraph

val create: int -> (int -> float * float) -> eucldigraph
(** `create n f` crée un graphe avec `n` sommets, le sommet `i` étant
    à la position `f i` ; aucun arc *)

val size: eucldigraph -> int
(** le nombre de sommets *)

val pos: eucldigraph -> int -> float * float
(** la position d'un sommet *)

val add_edge: eucldigraph -> int -> int -> unit

val has_edge: eucldigraph -> int -> int -> bool

val weight: eucldigraph -> int -> int -> float
(** le poids d'un arc i.e. la distance entre les deux sommets *)

val succ: eucldigraph -> int -> int list
(** voisinnage *)

val wdigraph: eucldigraph -> Wdigraph.wdigraph

val load: string -> eucldigraph

val print: Format.formatter -> eucldigraph -> unit

val distance: float * float -> float * float -> float
(** la distance euclidienne entre deux points *)

