(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

(** Solutions d'exercices *)

(* nombre de mots reconnus de longueur k *)
let count a k =
  if not (Auto.is_det a) then invalid_arg "count";
  let n = Auto.size a in
  let m = Matrix.make n 0 in
  for i = 0 to n - 1 do
    let trans = Auto.all_trans_opt a i in
    List.iter (fun (_, j) -> m.(i).(j) <- m.(i).(j) + 1) trans
  done;
  let v = Array.init n (fun i -> Math.iverson (Auto.is_final a i)) in
  (Matrix.power_apply m k v).(0)

let rec list_first f = function
  | [] -> None
  | x :: l -> ( match f x with None -> list_first f l | Some _ as r -> r)

let rec int_first lo hi f =
  if lo >= hi then None
  else match f lo with None -> int_first (lo + 1) hi f | Some _ as r -> r

(* un mot reconnu de longueur k

   par mémoïsation ou programmation dynamique, on calcule plus généralement
   f(s,j) un mot reconnu de longueur j<=k à partir de l'état s, et la
   solution est f(0,k) *)
let some_word a k =
  let memo = Hashtbl.create 16 in
  let rec f s j =
    assert (0 <= j && j <= k);
    try Hashtbl.find memo (s, j)
    with Not_found ->
      let w =
        if j = 0 then if Auto.is_final a s then Some "" else None
        else
          let cons c = function
            | None -> None
            | Some s -> Some (String.make 1 (Char.chr c) ^ s)
          in
          int_first 0 256 (fun c ->
              list_first
                (fun s' -> cons c (f s' (j - 1)))
                (Auto.trans a s (Char.chr c)))
      in
      Hashtbl.add memo (s, j) w;
      w
  in
  match f 0 k with None -> raise Not_found | Some s -> s
