
(* Graphe non orienté.

   Les sommets sont les entiers 0,1,...,size-1.
*)

type graph

val create: int -> graph

val size: graph -> int

val has_edge: graph -> int -> int -> bool

val add_edge: graph -> int -> int -> unit

val succ: graph -> int -> int list

type edge = int * int

val edges: graph -> edge list

val dfs: graph -> int -> bool array

val has_cycle: graph -> int -> bool

val bfs: graph -> int -> int array

val is_spanning_tree: graph -> (int * int) list -> bool

val is_bipartite: graph -> bool

val is_matching: graph -> (int * int) list -> bool

val load: string -> graph

val print: Format.formatter -> graph -> unit

val digraph: graph -> Digraph.digraph
  (* le graphe orienté sous-jacent *)

(* Quelques graphes particuliers *)

val cycle: int -> graph
  (* Le cycle 0--1--2--...--(n-1). *)

val full: ?loop:bool -> int -> graph
  (* Graphe complet, avec ou sans boucles. *)

val grid: int -> int -> graph
  (* Grille n*m où le sommet (i,j) est l'entier i*m+j *)
