
type 'v hashtable = (string * 'v) list array

let create (m: int) : 'v hashtable =
  if m <= 0 then invalid_arg "create";
  Array.make m []

let hash (k: string) : int =
  let n = String.length k in
  let rec hash h i =
    if i = n then h else hash (31*h + Char.code k.[i]) (i+1) in
  hash 0 0

let bucket (h: 'v hashtable) (k: string) : int =
  let i = (hash k) land max_int in (* on garantit i >= 0 *)
  i mod (Array.length h)

let put (h: 'v hashtable) (k: string) (v: 'v) : unit =
  let rec update b = match b with
    | [] -> [k, v]
    | (k', _) :: b when k = k' -> (k, v) :: b
    | e :: b -> e :: update b in
  let i = bucket h k in
  h.(i) <- update h.(i)

let contains (h: 'v hashtable) (k: string) : bool =
  let rec lookup b = match b with
    | [] -> false
    | (k', _) :: b -> k = k' || lookup b in
  lookup h.(bucket h k)

let get (h: 'v hashtable) (k: string) : 'v =
  let rec lookup b = match b with
    | [] -> raise Not_found
    | (k', v) :: b -> if k = k' then v else lookup b in
  lookup h.(bucket h k)

let remove (h: 'v hashtable) (k: string) : unit =
  let rec update b = match b with
    | [] -> []
    | (k', _) :: b when k = k' -> b
    | e :: b -> e :: update b in
  let i = bucket h k in
  h.(i) <- update h.(i)

let iter f h =
  for i = 0 to Array.length h - 1 do List.iter f h.(i) done

let stat h =
  let hist = Hashtbl.create 16 in
  let m = ref 0 in
  let fill b =
    let n = List.length b in
    if n > !m then m := n;
    Hashtbl.replace hist n (try 1 + Hashtbl.find hist n with Not_found -> 1) in
  Array.iter fill h;
  Array.init (!m + 1) (fun i -> try Hashtbl.find hist i with Not_found -> 0)

open Format

let print p h =
  printf "[";
  let print_entry (k, v) = printf "%s -> %s; " k (p v) in
  let print_bucket = List.iter print_entry in
  for i = 0 to Array.length h - 1 do print_bucket h.(i) done;
  printf "[@."

