
(* Problème du sac à dos

   Entrée :
   - un tableau de couples (vi, wi) où vi est une valeur et wi un poids
   - un poids maximum wmax
   Sortie :
   - la valeur maximale d'un sous-ensemble des couples tel que
     la somme de leurs poids n'excède pas wmax

   Chaque élément ne peut être pris qu'une seule fois.
*)

(* version récursive, exponentielle *)
let knapsack_rec (items: (int * int) array) (wmax: int) : int =
  let rec aux i w =
    if i = 0 || w = 0 then
      0
    else
      let vi, wi = items.(i-1) in
      if wi > w then aux (i-1) w
      else max (vi + aux (i-1) (w-wi)) (aux (i-1) w)
  in
  aux (Array.length items) wmax

(* programmation dynamique par mémoïsation *)
let knapsack_memo (items: (int * int) array) (wmax: int) : int =
  let n = Array.length items in
  let tab = Array.make_matrix (n+1) (wmax+1) (-1) in
  let rec aux i w =
    if i = 0 || w = 0 then
      0
    else (
      if tab.(i).(w) = -1 then (
        let vi, wi = items.(i-1) in
        tab.(i).(w) <-
          if wi > w
          then aux (i-1) w
          else max (vi + aux (i-1) (w-wi)) (aux (i-1) w)
      );
      tab.(i).(w)
    ) in
  aux n wmax

(* programmation dynamique de bas en haut ; renvoie la matrice *)
let knapsack_dp (items: (int * int) array) (wmax: int) : int array array =
  let n = Array.length items in
  let tab = Array.make_matrix (n+1) (wmax+1) 0 in
  for i = 1 to n do
    for w = 1 to wmax do
      let vi, wi = items.(i-1) in
      tab.(i).(w) <-
        if wi > w
        then tab.(i-1).(w)
        else max (vi + tab.(i-1).(w-wi)) tab.(i-1).(w)
    done;
  done;
  tab

(* renvoie la solution c'est-à-dire la liste des indices des
   éléments sélectionnés *)
let knapsack_list (items: (int * int) array) (wmax: int) : int list =
  let n = Array.length items in
  let tab = knapsack_dp items wmax in
  let rec build i w =
    if i = 0 || w = 0 then
      []
    else
      let _, wi = items.(i - 1) in
      if tab.(i).(w) = tab.(i - 1).(w) then
        build (i - 1) w
      else
        i-1 :: build (i - 1) (w - wi)
  in
  build n wmax

