
(* Matrices.

   On se limite ici à des matrices carrées, non vides,
   à coefficients entiers. *)

type matrix = int array array

let size x =
  Array.length x

let make n v =
  Array.make_matrix n n v

let init n f =
  Array.init n (fun i -> Array.init n (fun j -> f i j))

let id n =
  init n (fun i j -> if i = j then 1 else 0)

(* code piéton *)
let mul x y =
  let n = Array.length x in
  assert (n > 0 && Array.length x.(0) = n);
  assert (Array.length y = n && Array.length y.(0) = n);
  let m = Array.make_matrix n n 0 in
  for i = 0 to n - 1 do
    for j = 0 to n - 1 do
      for k = 0 to n - 1 do
        m.(i).(j) <- m.(i).(j) + x.(i).(k) * y.(k).(j)
      done
    done
  done;
  m

let mul x y =
  let n = Array.length x in
  assert (n > 0 && Array.length x.(0) = n);
  assert (Array.length y = n && Array.length y.(0) = n);
  let f i j =
    let prod s k = x.(i).(k) * y.(k).(j) + s in
    Math.fold_interval 0 n 0 prod in
  init n f

let rec power x n =
  if n = 0 then
    id (size x)
  else
    let y = power x (n / 2) in
    if n mod 2 = 0 then mul y y else mul x (mul y y)

let apply x v =
  let n = size x in
  assert (Array.length v = n);
  let product i = Math.sum_interval 0 n (fun j -> x.(i).(j) * v.(j)) in
  Array.init n product

let print fmt m =
  let n = size m in
  Format.fprintf fmt "@[";
  for i = 0 to n - 1 do
    for j = 0 to n - 1 do
      Format.fprintf fmt "%3d" m.(i).(j);
      if j < n-1 then Format.fprintf fmt " "
    done;
    Format.fprintf fmt "@\n"
  done;
  Format.fprintf fmt "@]"

(* calcule directement x^n*v pour un vecteur v, ce qui diminue
   le nombre de multiplications de matrices par rapport à power *)
let rec power_apply x n v =
  if n = 0 then
    v
  else
    power_apply (mul x x) (n / 2)
      (if n mod 2 = 1 then apply x v else v)
