
(* Matrices.

   On se limite ici à des matrices carrées, non vides,
   à coefficients entiers. *)

type matrix = int array array

val make: int -> int -> matrix
  (* `make n v` renvoie la matrice de taille `n`
     où tous les éléments valent `v` *)

val init: int -> (int -> int -> int) -> matrix
  (* `init n f` renvoie la matrice de taille `n`
     où l'élément (i,j) vaut `f i j` *)

val size: matrix -> int
  (* le nombre de lignes (et de colonnes) de la matrice *)

val id: int -> matrix
  (* matrice identité *)

val mul: matrix -> matrix -> matrix
  (* multiplication *)

val power: matrix -> int -> matrix
  (* `power m k` calcule matrice `m^k` avec une exponentiation rapide *)

val apply: matrix -> int array -> int array
  (* application d'une matrice à un vecteur *)

val power_apply: matrix -> int -> int array -> int array
  (* `power_apply m k v` renvoie le vecteur `m^k v`,
      plus efficacement que `apply (power m k) v` *)

val print: Format.formatter -> matrix -> unit
