
(* un labyrinthe parfait construit à partir de union-find *)

open Union_find

let () = Random.self_init ()

let n = int_of_string Sys.argv.(1)
let uf = create (n * n)

type direction = H | V

let doors =
  let l = ref [] in
  for i = 0 to n-1 do for j = 0 to n-1 do
    if i < n-1 then l := (i, j, V) :: !l;
    if j < n-1 then l := (i, j, H) :: !l;
  done; done;
  Array.of_list !l

(* mélange de Knuth *)
let () =
  for j = 1 to Array.length doors - 1 do
    let k = Random.int (j + 1) in
    let t = doors.(j) in doors.(j) <- doors.(k); doors.(k) <- t
  done

let horiz = Array.make_matrix n n true
let vert  = Array.make_matrix n n true

let () =
  let f (i, j, hv) =
    let k = i * n + j in
    let k' = if hv = H then k + 1 else k + n in
    if find uf k <> find uf k' then (
      if hv = H then horiz.(i).(j) <- false else vert.(i).(j) <- false;
      union uf k k';
    ) in
  Array.iter f doors

open Format

let () =
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      printf "+";
      if i = 0 || vert.(i-1).(j) then printf "--" else printf "  ";
    done;
    printf "+@.";
    for j = 0 to n-1 do
      if i = 0 && j = 0 then printf ">" else
      if j = 0 || horiz.(i).(j-1) then printf "|" else printf " ";
      printf "  ";
    done;
    if i = n-1 then printf ">@." else printf "|@.";
  done;
  for j = 0 to n-1 do
    printf "+--";
  done;
  printf "+@."

