
(* Lecture des fichiers de la base MNIST de caractères manuscrits.

   Cf http://yann.lecun.com/exdb/mnist/
*)

open Format

let width = 28 and height = 28
let size = width * height

type image = {
  index: int;
  data: int array; (* taille 28x28 = 784, valeurs 0..255, par ligne *)
  label: int; (* 0..9 *)
}

let print_to_pgm file {data} =
  assert (Array.length data = size);
  let cout = open_out file in
  let fmt = formatter_of_out_channel cout in
  fprintf fmt "P2@.";
  fprintf fmt "%d %d@." width height;
  fprintf fmt "255@.";
  for i = 0 to height - 1 do
    for j = 0 to width - 1 do fprintf fmt "%d " data.(width * i + j) done;
    fprintf fmt "@."
  done;
  close_out cout

let load ?threshold ~lo ~hi file =
  let ci = open_in (file ^ "-images-idx3-ubyte") in
  assert (input_binary_int ci = 0x00000803);
  let ni = input_binary_int ci in
  if lo < 0 || lo > hi || hi > ni then invalid_arg "iter";
  assert (input_binary_int ci = height);
  assert (input_binary_int ci = width);
  let cl = open_in (file ^ "-labels-idx1-ubyte") in
  assert (input_binary_int cl = 0x00000801);
  assert (input_binary_int cl = ni);
  (* sauter les `lo` premiers ;
     pas terrible, mais `seek_in` est absolu, pas relatif *)
  ignore (really_input_string ci (lo * size));
  ignore (really_input_string cl (lo * 1));
  let r = Array.make (hi - lo) { index=0; data=[||]; label=0 } in
  let cap x = match threshold with
    | None -> x | Some t -> if x < t then 0 else 255 in
  for index = lo to hi - 1 do
    let data = Array.make size 0 in
    for i = 0 to height - 1 do
      for j = 0 to width - 1 do
        data.(width * i + j) <- cap (input_byte ci)
      done
    done;
    let label = input_byte cl in
    r.(index - lo) <- { index; data; label }
  done;
  r
