/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"
#include "bintree.h"
#include "morris.h"

/*
  parcours infixe d'un arbre binaire, en espace constant
  et en temps linéaire

     Joseph M. Morris
     Traversing binary trees simply and cheaply
     Information Processing Letters 9(5), 1979
*/

void morris_inplace_inorder(bintree *t) {
  while (t != NULL) {
    if (t->left == NULL) {
      // pas de sous-arbre gauche : on visite le noeud et on descend à droite
      putchar(t->value); putchar('.'); // <-- visiter t
      t = t->right;
    } else {
      // un sous-arbre gauche : on cherche le prédécesseur p et t
      bintree *p = t->left;
      while (p->right != NULL && p->right != t) {
        p = p->right;
      }
      if (p->right == NULL) {
        // on a trouvé p -> on le fait pointer sur t et on descend à gauche
        p->right = t;
        t = t->left;
      } else {
        // on est retombé sur t -> on avait déjà visité le sous-arbre gauche
        // on restaure p->right, on visite t et on descend à droite
        p->right = NULL;
        putchar(t->value); putchar('.'); // <-- visiter t
        t = t->right;
      }
    }
  }
}
