
(* Satisfiabilité par l'algorithme de Quine
   Cette version travaille sur des formules quelconques *)

open Fmla

(* Fonctions de construction, avec simplifications élémentaires *)
let smart_not f = match f with
  | True  -> False
  | False -> True
  | _     -> Not f
let smart_and f1 f2 = match f1, f2 with
  | False, _ | _, False -> False
  | True,  f | f, True  -> f
  | _, _                -> Bin (And, f1, f2)
let smart_or f1 f2 = match f1, f2 with
  | True,  _ | _, True  -> True
  | False, f | f, False -> f
  | _, _                -> Bin (Or, f1, f2)
let smart_imp f1 f2 = match f1, f2 with
  | False, _ | _, True -> True
  | True,  f           -> f
  | f, False           -> smart_not f
  | _, _               -> Bin (Imp, f1, f2)

(* Simplication d'une formule, par reconstruction *)
let rec simplify f = match f with
  | Var _ | True | False -> f
  | Not f             -> smart_not (simplify f)
  | Bin (And, f1, f2) -> smart_and (simplify f1) (simplify f2)
  | Bin (Or,  f1, f2) -> smart_or  (simplify f1) (simplify f2)
  | Bin (Imp, f1, f2) -> smart_imp (simplify f1) (simplify f2)

(* Exploration des valuations *)
let rec quine_sat f =
  match simplify f with
  | True  -> true
  | False -> false
  | f     -> let v = varmax f in
             quine_sat (subst v True f)
             || quine_sat (subst v False f)
