
(* Réduction polynomiale de 3SAT à 3COLOR

   Il s'agit de la réduction avec laquelle on démontre la NP-complétude
   de 3COLOR : elle n'est pas nécessairement faite pour être utilisée en
   pratique.
 *)

open Cnf
open Graph

let check_3color g c =
  List.for_all (fun (i, j) -> c.(i) <> c.(j)) (edges g)
  && Array.for_all (fun i -> c.(i) < 3) c

let sat2color nf =
  let nc = List.length nf.clauses in
  let g = create (3 + 2 * nf.nbvars + 6 * nc ) in
  let vt, vf, vb = 0, 1, nf.nbvars + 2 in
  let dummyx = -nf.nbvars - 1 in
  let vx x = vb + x in
  let vc k = 3 + 2 * nf.nbvars + 6 * k in
  let add_edges l = List.iter (fun (s, t) -> add_edge g s t) l in
  let add_triangle a b c = add_edges [(a, b); (a, c); (b, c)] in
  add_triangle vt vf vb;
  for x = 1 to nf.nbvars do add_triangle vb (vx x) (vx (-x)) done;
  for k = 0 to nc - 1 do
    let v = vc k in
    add_triangle v (v+1) (v+3);
    add_triangle (v+2) (v+4) (v+5);
    add_edges [(v+3,v+4); (v+5,vf); (v+5,vb)]
  done;
  let fill_clause cl =
    match List.length cl with
    | 3 -> cl
    | 2 -> dummyx :: cl
    | 1 -> dummyx :: dummyx :: cl
    | _ -> invalid_arg "3sat 2 3color"
  in
  let clause_inputs k cl =
    let v = vc k in
    List.iteri (fun i x -> add_edge g (vx x) (v + i))
      (fill_clause cl)
  in
  List.iteri clause_inputs nf.clauses;
  g

let sat_using_color nf =
  let decode color =
    let vt = 0 in
    let vx x =  nf.nbvars + 2 + x in
    let ct = color.(vt) in
    let v = Array.make (nf.nbvars + 1) false in
    for i = 1 to nf.nbvars do
      v.(i) <- color.(vx i) = ct
    done;
    v
  in
  let g = sat2color nf in
  decode (Color.color3 g)

