/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// graphe dont les sommets sont des chaînes

// note : on suppose les chaînes d'au plus 80 caractères

#include "prelude.h"
#include "list.h"
#include "hashtbl.h"
#include "graph.h"
#include "sgraph.h"

struct Sgraph {
  graph *g;
  hashtbl *index; // string -> int
  char* *name;   // int -> string
};

int sgraph_index(sgraph *g, char* s) {
  return hashtbl_get(g->index, s);
}
char *sgraph_name(sgraph *g, int i) {
  return g->name[i];
}
graph *sgraph_graph(sgraph *g) {
  return g->g;
}

// pas de strdup en C99
char *strdup_(char *s) {
  char *r = malloc(strlen(s) + 1);
  return strcpy(r, s);
}

sgraph *sgraph_load(char *file) {
  // on commence par lire tous les sommets, en les associant à des entiers
  FILE *f = fopen(file, "r");
  hashtbl *h = hashtbl_create(5003);
  char s1[81], s2[81];
  int v = 0;
  while (fscanf(f, "%s %s\n", s1, s2) != EOF) {
    if (!hashtbl_contains(h, s1)) {
      hashtbl_put(h, strdup_(s1), v++);
    }
    if (!hashtbl_contains(h, s2)) {
      hashtbl_put(h, strdup_(s2), v++);
    }
  }
  fclose(f);
  // puis on recommence, cette fois pour construire le graphe
  graph *g = graph_create(v);
  char **name = calloc(v, sizeof(char*));
  for (int i = 0; i < v; i++) name[i] = NULL;
  f = fopen(file, "r");
  while (fscanf(f, "%s %s\n", s1, s2) != EOF) {
    int i1 = hashtbl_get(h, s1), i2 = hashtbl_get(h, s2);
    assert(i1 >= 0 && i2 >= 0);
    // printf("s1=%s, s2=%s, i1=%d, i2=%d\n", s1, s2, i1, i2);
    if (name[i1] == NULL) { name[i1] = strdup_(s1); }
    if (name[i2] == NULL) { name[i2] = strdup_(s2); }
    graph_add_edge(g, i1, i2);
  }
  fclose(f);
  // enfin on alloue et on renvoie le résultat
  sgraph *sg = malloc(sizeof(struct Sgraph));
  sg->g = g;
  sg->index = h;
  sg->name = name;
  return sg;
}

void sgraph_print(sgraph *g) {
  int v = graph_size(g->g);
  printf("V = %d\n", v);
  for (int i = 0; i < v; i++) {
    for (list *l = graph_succ(g->g, i); l != NULL; l = l->next) {
      if (i < l->value) {
        printf("%s %s\n", sgraph_name(g, i), sgraph_name(g, l->value));
      }
    }
  }
}
