/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"

bool is_sorted(int a[], int n) {
  for (int i = 0; i < n-1; i++) {
    if (a[i] > a[i+1]) {
      return false;
    }
  }
  return true;
}

void swap(int a[], int i, int j) {
  int tmp = a[i];
  a[i] = a[j];
  a[j] = tmp;
}

void two_way_sort(int a[], int n) {
  int lo = 0, hi = n - 1;
  while (lo < hi) {
    if (a[lo] == 0) { lo++; }
    else if (a[hi] == 1) { hi--; }
    else { swap(a, lo++, hi--); }
  }
}

void knuth_shuffle(int a[], int n) {
  for (int i = 1; i < n; i++) {
    swap(a, i, rand() % (i+1));
  }
}

// drapeau hollandais
// on suppose ici que le tableau contient uniquement les valeurs 0, 1, 2
//    0      b       i       r     n
//   +------+-------+-------+-----+
//   |   0  |   1   | ????? |  2  |
//   +------+-------+-------+-----+
void dutch_flag(int a[], int n) {
  int b = 0, i = 0, r = n;
  while (i < r) {
    if (a[i] == 0) { swap(a, b++, i++); }
    else if (a[i] == 1) { i++; }
    else { swap(a, --r, i); }
  }
}

// tri par sélection

void selection_sort(int a[], int n) {
  for (int i = 0; i < n - 1; i++) {
    int j_min = i;
    for (int j = i + 1; j < n; j++) {
      if (a[j] < a[j_min]) { j_min = j; }
    }
    int v = a[j_min];
    a[j_min] = a[i];
    a[i] = v;
  }
}

typedef struct Data {
  int key;
  char *contents;
} data;

void swap_data(data a[], int i, int j) {
  data tmp = a[i];
  a[i] = a[j];
  a[j] = tmp;
}

int select_min(data a[], int lo, int hi) {
  assert (lo < hi);
  int i_min = lo;
  for (int i = lo + 1; i < hi; i++) {
    if (a[i].key < a[i_min].key) { i_min = i; }
  }
  return i_min;
}

void selection_sort_data(data a[], int n) {
  for (int i = 0; i < n; i++) {
    int j = select_min(a, i, n);
    swap_data(a, i, j);
  }
}

// tri par insertion

void insertion_sort(int a[], int n) {
  for (int i = 1; i < n; i++) {
    // invariant: a[0..i[ est trié
    int v = a[i];
    int j = i;
    while (j > 0 && a[j-1] > v) {
      a[j] = a[j-1];
      j--;
    }
    a[j] = v;
  }
}

// renvoie la position immédiatement à droite de v dans a[lo..hi[
// (en supposant a[lo..hi[ trié par ordre croissant)
int binary_search_right(int v, int a[], int lo, int hi) {
  assert(lo <= hi);
  while (lo < hi) {
    int mid = lo + (hi - lo) / 2;
    if (a[mid] <= v) lo = mid+1; else hi = mid;
  }
  return hi;
}

void binary_sort(int a[], int n) {
  for (int i = 1; i < n; i++) {
    // invariant: a[0..i[ est trié
    int v = a[i];
    int m = binary_search_right(v, a, 0, i);
    for (int j = i; j > m; j--) {
      a[j] = a[j-1];
    }
    a[m] = v;
  }
}

// tri par tas

void move_down(int a[], int i, int x, int n) {
  while (true) {
    int j = 2 * i + 1;
    if (j >= n) break;
    if (j + 1 < n && a[j] < a[j + 1]) { j++; }
    if (a[j] <= x) break;
    a[i] = a[j];
    i = j;
  }
  a[i] = x;
}

void heapsort_(int a[], int n) {
  for (int k = n / 2 - 1; k >= 0; k--)
    move_down(a, k, a[k], n);
  for (int k = n - 1; k >= 1; k--) {
    int v = a[k];
    a[k] = a[0];
    move_down(a, 0, v, k);
  }
}

// quicksort, avec une partition en 3 (c'est le drapeau hollandais)
// et un mélange pour démarrer

//    l      lo      i       hi     r
//   +------+-------+-------+------+
//   |  <p  |  =p   | ????? |  >p  |
//   +------+-------+-------+------+

void quickrec0(int a[], int l, int r) { // trie uniquement a[l..r[
  if (r - l <= 1) return;
  int p = a[l], lo = l, hi = r;
  for (int i = l+1; i < hi; ) {
    if (a[i] == p) {
      i++;
    } else if (a[i] < p) {
      swap(a, i++, lo++);
    } else { // p < a[i]
      swap(a, i, --hi);
    }
  }
  quickrec0(a, l, lo);
  quickrec0(a, hi, r);
}

// encore mieux : sans faire déborder la pile

void quickrec(int a[], int l, int r) { // trie uniquement a[l..r[
  if (r - l <= 1) return;
  int p = a[l], lo = l, hi = r;
  for (int i = l+1; i < hi; ) {
    if (a[i] < p) {
      swap(a, i++, lo++);
    } else if (a[i] == p) {
      i++;
    } else { // a[i] > p
      swap(a, i, --hi);
    }
  }
  if (lo - l < r - hi) {
    quickrec(a, l, lo);
    quickrec(a, hi, r);
  } else {
    quickrec(a, hi, r);
    quickrec(a, l, lo);
  }
}

void quicksort(int a[], int n) {
  knuth_shuffle(a, n);
  quickrec(a, 0, n);
}

// quick select (a.k.a. Hoare's find)

int quickselectrec(int k, int a[], int l, int r) {
  // cherche l'élément de rang (k-l) dans a[l..r[
  assert (l <= k && k < r);
  // if (r == l+1) return a[l]; // pas nécessaire
  int p = a[l], lo = l, hi = r;
  for (int i = l+1; i < hi; ) {
    if (a[i] == p) {
      i++;
    } else if (a[i] < p) {
      swap(a, i++, lo++);
    } else { // p < a[i]
      swap(a, i, --hi);
    }
  }
  if      (k < lo)   return quickselectrec(k, a, l, lo);
  else if (k < hi)   return a[k];
  else /* k >= hi */ return quickselectrec(k, a, hi, r);
}

int quickselect(int k, int a[], int n) {
  assert (0 <= k && k < n);
  return quickselectrec(k, a, 0, n);
}

int slowselect(int k, int a[], int n) {
  for (int i = 0; i <= k; i++) {
    int j_min = i;
    for (int j = i + 1; j < n; j++) {
      if (a[j] < a[j_min]) { j_min = j; }
    }
    swap(a, i, j_min);
  }
  return a[k];
}

// version alternative de find, avec bipartition

int findrec(int k, int a[], int l, int r) {
  // cherche l'élément de rang (k-l) dans a[l..r[
  assert (l <= k && k < r);
  int p = a[l], lo = l+1, hi = r;
  while (lo < hi) {
    if (p < a[lo]) { swap(a, lo, --hi); }
    else { lo++; }
  }
  swap(a, l, lo-1);
  if      (k < lo-1)   return findrec(k, a, l, lo-1);
  else if (k >= lo)    return findrec(k, a, lo, r);
  else /* k == lo-1 */ return p;
}

int find(int k, int a[], int n) {
  assert (0 <= k && k < n);
  return findrec(k, a, 0, n);
}

// différentes variantes de bipartition
int bipartition(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n;
  while (lo < hi) {
    if (p < a[lo]) { swap(a, lo, --hi); }
    else { lo++; }
  }
  swap(a, 0, lo - 1);
  return lo - 1;
}

int bipartition2(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n - 1;
  while (lo <= hi) {
    if (a[lo] <= p) { lo++; }
    else if (a[hi] >= p) { hi--; }
    else { swap(a, lo++, hi--); }
  }
  swap(a, 0, lo);
  return lo;
}

int bipartition3(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n-1;
  while (lo <= hi) {
    while (lo < n && a[lo] <= p) { lo++; }
    while (hi > 0 && a[hi] >= p) { hi--; }
    if (lo < hi) { swap(a, lo++, hi--); }
  }
  swap(a, 0, lo);
  return lo;
}

// tri fusion

// fusionne a1[l..m[ et a1[m..r[ dans a2[l..r[
void merge(int a1[], int a2[], int l, int m, int r) {
  int i = l, j = m;
  for (int k = l; k < r; k++)
    if (i < m && (j == r || a1[i] <= a1[j]))
      a2[k] = a1[i++];
    else
      a2[k] = a1[j++];
}

void mergesortrec(int a[], int tmp[], int l, int r) {
  if (r - l <= 1) return;
  int m = l + (r - l) / 2;
  mergesortrec(a, tmp, l, m);
  mergesortrec(a, tmp, m, r);
  if (a[m-1] <= a[m]) return; // optim
  for (int i = l; i < r; i++) tmp[i] = a[i];
  merge(tmp, a, l, m, r);
}

void mergesort_(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  mergesortrec(a, tmp, 0, n);
  free(tmp);
}

int merge_count(int a1[], int a2[], int l, int m, int r) {
  int i = l, j = m;
  int c = 0;
  for (int k = l; k < r; k++)
    if (i < m && (j == r || a1[i] <= a1[j]))
      a2[k] = a1[i++];
    else {
      a2[k] = a1[j++];
      c += m - i;
    }
  return c;
}

int mergesortrec_count(int a[], int tmp[], int l, int r) {
  if (r - l <= 1) return 0;
  int m = l + (r - l) / 2;
  int cleft = mergesortrec_count(a, tmp, l, m);
  int cright = mergesortrec_count(a, tmp, m, r);
  if (a[m-1] <= a[m]) return cleft + cright; // optim
  for (int i = l; i < r; i++) tmp[i] = a[i];
  int cmerge = merge_count(tmp, a, l, m, r);
  return cleft + cright + cmerge;
}

int mergesort_count(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  int c = mergesortrec_count(a, tmp, 0, n);
  free(tmp);
  return c;
}


int min(int x, int y) { return x < y ? x : y; }

void bottom_up_mergesort(int a[], int n) {
  int *tmp = calloc(n, sizeof(int));
  for (int len = 1; len < n; len *= 2) {
    for (int i = 0; i < n; i++) tmp[i] = a[i];
    for (int k = 0; k < n - len; k += 2 * len) {
      merge(tmp, a, k, k + len, min(k + 2 * len, n));
    }
  }
  free(tmp);
}

int find_run(int a[], int lo, int n) {
  while (++lo < n && a[lo-1] <= a[lo]);
  return lo;
}

void natural_mergesort(int a[], int n) {
  if (n <= 1) return;
  int *tmp = calloc(n, sizeof(int));
  while (true) {
    for (int lo = 0; lo < n-1; ) {
      // le premier run a[lo..mid[
      int mid = find_run(a, lo, n);
      if (mid == n) { if (lo == 0) return; break; }
      // le second run a[mid..hi[
      int hi = find_run(a, mid, n);
      // et on les fusionne
      for (int i = lo; i < hi; i++) tmp[i] = a[i];
      merge(tmp, a, lo, mid, hi);
      lo = hi;
    }
  }
}
