
(* Arbres préfixes (trie)

   On représente ici des ensembles de mots (string) mais on pourrait
   facilement faire

   - des dictionnaires plutôt que des ensembles ;

   - un alphabet plus large et/ou non nécessairement continu, avec un
   branchement réalisé autrement qu'avec un tableau ;

   - une structure immuable plutôt que mutable.  *)

type trie = {
  mutable present: bool;
         branches: (char, trie) Hashtbl.t;
}

let create () =
  { present = false; branches = Hashtbl.create 8; }

let mem t s =
  let n = String.length s in
  let rec find t i =
    if i = n then
      t.present
    else
      try find (Hashtbl.find t.branches s.[i]) (i + 1)
      with Not_found -> false in
  find t 0

let add t s =
  let n = String.length s in
  let rec add t i =
    if i = n then
      t.present <- true
    else
      let b =
        try
          Hashtbl.find t.branches s.[i]
        with Not_found ->
          let b = create () in
          Hashtbl.add t.branches s.[i] b;
          b
      in
      add b (i+1)
  in
  add t 0

(* Note: on prend soin de supprimer les branches vides (en remontant) *)

let is_empty t =
  not t.present && Hashtbl.length t.branches = 0

let remove t s =
  let n = String.length s in
  let rec remove t i =
    if i = n then
      t.present <- false
    else
      try
        let b = Hashtbl.find t.branches s.[i] in
        remove b (i+1);
        if is_empty b then Hashtbl.remove t.branches s.[i]
      with Not_found -> ()
  in
  remove t 0

