
(* Réduction de SAT vers 3-SAT par la transformation de Tseitin *)

open Fmla
open Cnf

let tseitin (f: fmla) : nf =
  let nv = ref (varmax f) in
  let new_var () = incr nv; !nv in
  let rec mk_clauses = function
    | Var z -> z, []
    | True  -> let x = new_var () in x, [[x]]
    | False -> let x = new_var () in x, [[-x]]
    | Not f -> let x = new_var () in
               let y, cls = mk_clauses f in
               x, [-x; -y] :: [x; y] :: cls
    | Bin (op, f1, f2) ->
       let x = new_var () in
       let y1, cls1 = mk_clauses f1 in
       let y2, cls2 = mk_clauses f2 in
       let cls = match op with
         | And -> [-x; y1] :: [-x; y2] :: [x; -y1; -y2] :: cls1 @ cls2
         | Or  -> [-x; y1; y2] :: [x; -y1] :: [x; -y2] :: cls1 @ cls2
         | Imp -> [-x; -y1; y2] :: [x; y1] :: [x; -y2] :: cls1 @ cls2
       in
       x, cls
  in
  let x, cls = mk_clauses f in
  { kind = CNF; nbvars = !nv; clauses = [x] :: cls }


(* Si l'on connaissait une solution polynomiale au problème 3SAT,
   on en déduirait une solution polynomiale au problème SAT. *)
let polynomial_3sat (f: nf) : bool =
  failwith "what did you expect?"

let polynomial_sat (f: fmla) : bool =
  let nf = tseitin f in
  polynomial_3sat nf
(* À défaut, on peut utiliser un algorithme DPLL sur la formule 3SAT
   produite par la transformation de Tseitin *)
