
type 'a vector = {
  mutable data: 'a array;
  mutable size: int; (* 0 <= size <= Array.length data *)
  default: 'a;
}
(* invariant : la partie inutilisée, c'est-à-dire data[size..[, est
   remplie avec `default` *)

let create cap def =
  { data = Array.make cap def;
    size = 0;
    default = def; }

let size v =
  v.size

let get v i =
  if i < 0 || i >= v.size then invalid_arg "get";
  v.data.(i)

let set v i x =
  if i < 0 || i >= v.size then invalid_arg "set";
  v.data.(i) <- x

let resize v s =
  if s < 0 then invalid_arg "resize";
  let n = Array.length v.data in
  if s > n then begin
    let n' = max s (2 * n) in
    let old = v.data in
    v.data <- Array.make n' v.default;
    Array.blit old 0 v.data 0 v.size
  end else if s < v.size then begin
    (* pas de fuite mémoire *)
    Array.fill v.data s (v.size - s) v.default
  end;
  v.size <- s

let clear v =
  resize v 0

let iteri f v =
  for i = 0 to size v - 1 do f i (get v i) done

(* on peut notamment l'utiliser comme une pile *)

let push v x =
  let n = v.size in
  resize v (n + 1);
  set v n x

let pop v =
  if v.size = 0 then invalid_arg "pop";
  let n = v.size - 1 in
  let x = get v n in
  resize v n;
  x

let top v =
  if v.size = 0 then invalid_arg "top";
  get v (v.size - 1)
