/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// file réalisée avec un tableau redimensionnable (circulaire)
// (exercice)

#include "prelude.h"

#include "vector.h"
#include "vqueue.h"

struct VQueue {
  int size;  // 0 <= size <= capacity = vector_size(data)
  int front; // 0 <= front < capacity
  vector *data;
};

vqueue *vqueue_create(int capacity) {
  vqueue *q = malloc(sizeof(struct VQueue));
  q->size = 0;
  q->front = 0;
  q->data = vector_create(capacity);
  vector_resize(q->data, capacity);
  return q;
}

int vqueue_size(vqueue *q) {
  return q->size;
}

bool vqueue_is_empty(vqueue *q) {
  return vqueue_size(q) == 0;
}

bool vqueue_is_full(vqueue *q) {
  return q->size == vector_size(q->data);
}

void vqueue_enqueue(vqueue *q, int x) {
  if (vqueue_is_full(q)) {
    int c = vector_size(q->data);
    if (q->front == 0) {
      // cas simple où il suffit d'ajouter une place à la fin
      vector_resize(q->data, c + 1);
    } else {
      vector_resize(q->data, c + q->front);
      for (int i = 0; i < q->front; i++) {
        vector_set(q->data, c + i, vector_get(q->data, i));
      }
    }
  }
  int c = vector_size(q->data);
  int i = q->front + q->size;
  if (i >= c) i -= c;
  vector_set(q->data, i, x);
  q->size++;
}

int vqueue_peek(vqueue *q) {
  assert(!vqueue_is_empty(q));
  return vector_get(q->data, q->front);
}

int vqueue_dequeue(vqueue *q) {
  assert(!vqueue_is_empty(q));
  int x = vector_get(q->data, q->front);
  q->front++;
  if (q->front == vector_size(q->data)) q->front = 0;
  q->size--;
  return x;
}

void vqueue_delete(vqueue *q) {
  vector_delete(q->data);
  free(q);
}
