
(* Graphes orientés pondérés, par listes d'adjacence *)

type wdigraph = (int * float) list array

let create n =
  Array.make n []

let size g =
  Array.length g

let has_edge g u v =
  List.mem_assoc v g.(u)

type edge = float * int * int

let add_edge g (d, u, v) =
  assert (not (has_edge g u v));
  g.(u) <- (v, d) :: g.(u)

let weight g u v =
  assert (has_edge g u v);
  List.assoc v g.(u)

let succ g u =
  g.(u)

let edges g =
  let l = ref [] in
  for i = 0 to size g - 1 do
    List.iter (fun (j, d) -> l := (d, i, j) :: !l) g.(i)
  done;
  !l

open Scanf

let load file =
  let c = Scanning.open_in file in
  let n = bscanf c "%d\n" (fun n -> n) in
  let g = create n in
  let e = bscanf c "%d\n" (fun e -> e) in
  for _ = 1 to e do
    bscanf c "%d %d %f\n" (fun u v d -> add_edge g (d,u,v))
  done;
  g

open Format

let print fmt g =
  fprintf fmt "%d@\n" (size g);
  let e = Array.fold_left (List.fold_left (fun e _ -> e+1)) 0 g in
  fprintf fmt "%d@\n" e;
  for i = 0 to size g - 1 do
    List.iter (fun (j, d) -> fprintf fmt "%d %d %f@\n" i j d) g.(i)
  done

